package be.iminds.ilabt.jfed.experimenter_gui;

import be.iminds.ilabt.jfed.BaseGuiModule;
import be.iminds.ilabt.jfed.UncaughtExceptionLogger;
import be.iminds.ilabt.jfed.experimenter_gui.ext_requests.EnvironmentParameterHandler;
import be.iminds.ilabt.jfed.experimenter_gui.ext_requests.ExternalUserRequestHandler;
import be.iminds.ilabt.jfed.experimenter_gui.ext_requests.FirstLoginParametersHandler;
import be.iminds.ilabt.jfed.experimenter_gui.login.UserLoginController;
import be.iminds.ilabt.jfed.experimenter_gui.preferences.PreferencesDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.slice.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentViewController;
import be.iminds.ilabt.jfed.experimenter_gui.ui.AcceleratorService;
import be.iminds.ilabt.jfed.experimenter_gui.util.MacOpenURIHandler;
import be.iminds.ilabt.jfed.highlevel.HighLevelModule;
import be.iminds.ilabt.jfed.highlevel.LowLevelModule;
import be.iminds.ilabt.jfed.highlevel.UserLoginModelModule;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.logging.JFedLoggerManager;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.UserLoginModelManager;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.preferences.JFedPreferencesApplier;
import be.iminds.ilabt.jfed.ssh_terminal_tool.putty.PageantHelper;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.UserSshKeyInfo;
import be.iminds.ilabt.jfed.ssh_terminal_tool.unix.SshAgentHelper;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.ui.javafx.util.ExtraInfoCallBackGUIs;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.JFedUtils;
import be.iminds.ilabt.jfed.util.OSDetector;
import com.cathive.fx.guice.GuiceApplication;
import com.cathive.fx.guice.GuiceFXMLLoader;
import com.google.inject.Module;
import com.google.inject.grapher.graphviz.GraphvizModule;
import javafx.application.Platform;
import javafx.collections.ObservableMap;
import javafx.scene.Scene;
import javafx.scene.control.ButtonType;
import javafx.scene.image.Image;
import javafx.stage.Stage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Provider;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * User: twalcari
 * Date: 10/31/13
 * Time: 6:17 PM
 */
public class Launch extends GuiceApplication implements Provider<Stage> {
    /* solution for MAC OS HeadlessException bug. See https://ibcn-jira.intec.ugent.be/browse/FEDIBBTDEV-239 */
    static {
        System.setProperty("java.awt.headless", "false");
    }

    private static final Logger LOG = LoggerFactory.getLogger(Launch.class);
    private static final String SCHEME_FILE = "file";


    @Inject
    private JFedLoggerManager jFedLoggerManager;
    @Inject
    private JFedGuiPreferences jFedPreferences;
    @Inject
    private JFedPreferencesApplier jFedPreferencesApplier;
    @Inject
    private UserLoginModelManager userLoginModelManager;
    @Inject
    private ObservableMap<GeniUrn, ExperimentViewController> experimentViewControllers;
    @Inject
    private TaskThread tt;
    @Inject
    private GuiceFXMLLoader loader;
    @Inject
    private PageantHelper pageantHelper;

    @Inject
    private PreferencesDialogFactory preferencesDialogFactory;

    private Stage primaryStage;
    @Inject
    private AcceleratorService acceleratorService;

    @Inject
    private FirstLoginParametersHandler firstLoginParametersHandler;

    @Inject
    private ExternalUserRequestHandler externalUserRequestHandler;

    @Inject
    private EnvironmentParameterHandler environmentParameterHandler;

    @Inject
    private Provider<MacOpenURIHandler> macOpenURIHandlerProvider;

    //keep reference to prevent it from being Garbage-collected
    private MacOpenURIHandler macOpenURIHandler = null;

    public static void main(String[] args) {
        launch(args);
    }

    @Override
    public void start(final Stage stage) throws Exception {
        this.primaryStage = stage;

        //catch all exceptions which aren't caught anywhere else
        Thread.setDefaultUncaughtExceptionHandler(new UncaughtExceptionLogger());

        try {
            //check for filesystem access
            if (JFedUtils.getUserDataDirectory() == null) {

                JFDialogs.create().owner(stage).message("Please verify your security settings and try again.\n\n" +
                        "If you are trying to run jFed through Safari on OS X, you need to change your Safari security settings.\n" +
                        "(cfr. http://doc.fed4fire.eu/firstexperiment.html#start-up-jfed)")
                        .masthead("jFed was unable to access the filesystem.")
                        .title("Fatal Error").showError();

                exit();
                return;
            }

            //This should already be done (when the injected object was created), but just to be sure, we again update the Loggers here
            jFedLoggerManager.updateLoggers();

            //register Mac URI handler when on Mac
            if (OSDetector.findOS() == OSDetector.OS.MAC) {
                macOpenURIHandler = macOpenURIHandlerProvider.get();
            }

            //process startup parameters on all platforms
            Map<String, String> parameters = getParameters().getNamed();
            environmentParameterHandler.handleParameters(parameters);

            ExtraInfoCallBackGUIs.prepareExtraInfoCallback(stage.getOwner());

            FirstLoginParametersHandler.LoginResult loginResult =
                    firstLoginParametersHandler.handleParameters(parameters);

            if (!loginResult.isLoggedIn()) {
                if (!loginResult.isLoginDefined()) {
                    userLoginModelManager.load();
                }

                //does not wait! UserLoginControllerListener can be used to see if login window is closed
                showUserLogin(stage);
            } else {
                onSuccessfulLogin(stage);
            }
        } catch (Throwable ex) {
            LOG.error("Fatal error in jFed Experimenter Main Thread: " + ex.getMessage(), ex);
            JFDialogs.create()
                    .owner(stage)
                    .message("jFed experimenter will terminate.")
                    .masthead("A fatal error occurred!")
                    .title("Fatal Error")
                    .showException(ex);

            throw ex;
        }
    }

    private void onSuccessfulLogin(Stage stage) {
        try {
            //show information dialog about public/private keys on first run of jFed

            if (jFedPreferences.getInt(GuiPreferenceKey.PREF_FIRSTRUN_VERSION) < JFedPreferences.CURRENT_FIRSTRUN_VERSION) {

                if (OSDetector.findOS() == OSDetector.OS.WIN) {


                    JFDialogs.create()
                            .owner(stage)
                            .message("The default preferences should be okay. If you have trouble running experiments or connecting to nodes, consider to turn on 'Proxy' in the jFed Proxy preferences.\n\n" +
                                    "Next you will see the preferences window, you can always revisit this by clicking the preferences button.\n\n" +
                                    "Please verify that jFed has found the PuTTY SSH software (you should see 4 green checkmarks). If not install first PuTTY (full suite) or point jFed to the right PuTTY install directory.")
                            .masthead("Initial configuration")
                            .title("Welcome to the jFed Experimenter Tool!")
                            .showInformation();


                } else {

                    JFDialogs.create()
                            .owner(stage)
                            .message("The default preferences should be okay. If you have trouble running experiments or connecting to nodes, consider to turn on 'Proxy' in the jFed Proxy preferences.\n\n" +
                                    "Next you will see the preferences window, you can always revisit this by clicking the preferences button.")
                            .masthead("Initial configuration")
                            .title("Welcome to the jFed Experimenter Tool!")
                            .showInformation();

                }

                preferencesDialogFactory.showDialog();

                jFedPreferences.setInt(GuiPreferenceKey.PREF_FIRSTRUN_VERSION, JFedPreferences.CURRENT_FIRSTRUN_VERSION);
            }

            jFedPreferencesApplier.applyPreferences();

            showExperimenterGui(stage);
        } catch (Throwable ex) {
            LOG.error("Fatal error in jFed Experimenter Main Thread: " + ex.getMessage(), ex);
            JFDialogs.create()
                    .owner(stage)
                    .message("jFed experimenter will terminate.")
                    .masthead("A fatal error occurred!")
                    .title("Fatal Error")
                    .showException(ex);

            throw ex;
        }
    }

    private void registerKeysWithPageant() {
        if (jFedPreferences.getBoolean(GuiPreferenceKey.PREF_PUTTY_PAGEANT, false)) {
            List<SshKeyInfo> keys = new ArrayList<>();

            SshKeyInfo sshKeyInfo = jFedPreferences.getOverriddenSshKeyInfo();
            if (sshKeyInfo == null && userLoginModelManager.isUserLoggedIn()) {
                GeniUser geniUser = userLoginModelManager.getLoggedInGeniUser();
                sshKeyInfo = new UserSshKeyInfo(geniUser);
            }
            //PageantHelper.registerKey(sshKeyInfo);
            keys.add(sshKeyInfo);

            if (!jFedPreferences.getString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH).equalsIgnoreCase("NEVER")) {
                SshKeyInfo proxySshKeyInfo = jFedPreferences.getOverriddenProxySshKeyInfo();
                if (proxySshKeyInfo == null && userLoginModelManager.isUserLoggedIn()) {
                    GeniUser geniUser = userLoginModelManager.getLoggedInGeniUser();
                    proxySshKeyInfo = new UserSshKeyInfo(geniUser);
                }
                //PageantHelper.registerKey(proxySshKeyInfo);
                keys.add(proxySshKeyInfo);
            }

            pageantHelper.registerKeys(keys);
        }
    }

    /**
     * doesn't wait
     */
    public void showUserLogin(final Stage parentStage) {

        UserLoginController.UserLoginControllerListener userLoginControllerListener = successfulLogin -> {
            if (successfulLogin)
                onSuccessfulLogin(parentStage);
            else {
                exit();
            }
        };

        try {
            URL location = UserLoginController.class.getResource("UserLogin.fxml");

            GuiceFXMLLoader.Result result = loader.load(location);

            final UserLoginController userLoginController = result.getController();
            userLoginController.addListener(userLoginControllerListener);

            Stage userLoginStage = new Stage();
            userLoginController.setupStage(userLoginStage);

            userLoginStage.setTitle("jFed login");
            userLoginStage.setResizable(false);
            userLoginStage.setScene(new Scene(result.getRoot()));
            userLoginStage.getIcons().addAll(new Image(UserLoginController.class.getResource("/images/jfed-logo_16.png").toExternalForm())
                    , new Image(UserLoginController.class.getResource("/images/jfed-logo_32.png").toExternalForm())
                    , new Image(UserLoginController.class.getResource("/images/jfed-logo_64.png").toExternalForm()));


            //don't wait!
            userLoginStage.show();
        } catch (Exception e) {
            throw new RuntimeException("Something went wrong showing the User Login Screen: " + e.getMessage(), e);
        }
    }


    private boolean isExperimentActive(Experiment experiment) {
        if (experiment.getSlice() == null)
            return false;

        for (Sliver sliver : experiment.getSlice().getSlivers()) {
            switch (sliver.getStatus().getGlobalStatus()) {
                case READY:
                case UNKNOWN:
                case CHANGING:
                    return true;
            }
        }
        return false;
    }


    public void showExperimenterGui(Stage stage) {
        try {
            GuiceFXMLLoader.Result result = loader.load(getClass().getResource("ExperimenterGUI.fxml"));
            Scene scene = new Scene(result.getRoot());

            stage.setScene(scene);
            stage.setMinHeight(600);
            stage.setMinWidth(800);

            stage.setTitle("jFed Experimenter Toolkit");
            stage.getIcons().addAll(new Image(Launch.class.getResource("/images/jfed-logo_16.png").toExternalForm())
                    , new Image(Launch.class.getResource("/images/jfed-logo_32.png").toExternalForm())
                    , new Image(Launch.class.getResource("/images/jfed-logo_64.png").toExternalForm()));

            stage.setOnCloseRequest(windowEvent -> {
                boolean hasActiveSlices = experimentViewControllers.values().stream()
                        .map(ExperimentViewController::getExperiment)
                        .anyMatch(this::isExperimentActive);

                if (hasActiveSlices) {
                    //show a warning
                    JFDialogs.create()
                            .message("There are currently resources in use.\nAre you sure you want to exit?")
                            .title("Exit while resources in use").buttonTypes(ButtonType.YES, ButtonType.NO, ButtonType.CANCEL)
                            .showWarning()
                            .filter(response -> response != ButtonType.YES)
                            .ifPresent(response -> windowEvent.consume());

                    if (windowEvent.isConsumed()) {
                        return;
                    }

                }

                exit();
            });

            //Called just after the Window is shown.
            stage.setOnShown(windowEvent -> registerKeysWithPageant());

            acceleratorService.registerOnScene(scene);

            externalUserRequestHandler.handleParameters(getParameters().getNamed());

            stage.show();
        } catch (IOException ex) {
            LOG.error("Error while loading ExperimenterGUI", ex);
            throw new RuntimeException("Error while loading ExperimenterGUI", ex);
        }
    }

    public void exit() {
        LOG.info("Close was requested. Trying to terminate");
        tt.requestStop();

        //stop any running ssh-agent
        SshAgentHelper.stopSshAgent();

        Platform.exit();
        System.exit(0);
    }

    @Override
    public void init(List<Module> modules) throws Exception {
        modules.add(new LowLevelModule());
        modules.add(new HighLevelModule());
        modules.add(new UserLoginModelModule());
        modules.add(new ExperimenterModule());
        modules.add(new BaseGuiModule());
        modules.add(new GraphvizModule());
    }

    @Override
    public Stage get() {
        return primaryStage;
    }
}
